﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes the quality of schemas.
  /// </summary>
  [Item("BestIndividualQualityAnalyzer", "An operator that analyzes broken inheritance of building blocks.")]
  [StorableClass]
  public sealed class BestIndividualQualityAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemataParameterName = "BestIndividualSchemata";
    private const string SchemaQualityMatrixParameterName = "SchemaQualityMatrix";
    private const string QualitiesParameterName = "Qualities";
    private const string DistanceMatrixParameterName = "DistanceMatrix";
    private const string GenerationsParameterName = "Generations";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DistanceMatrix>)Parameters[DistanceMatrixParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.ActualValue; }
    }
    public HeuristicLab.Analysis.DataTable Qualities {
      get { return ((HeuristicLab.Analysis.DataTable)ResultsParameter.ActualValue[QualitiesParameterName].Value); }
    }
    #endregion

    [StorableConstructor]
    private BestIndividualQualityAnalyzer(bool deserializing) : base(deserializing) { }
    private BestIndividualQualityAnalyzer(BestIndividualQualityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public BestIndividualQualityAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should are stored."));
      Parameters.Add(new LookupParameter<DistanceMatrix>(DistanceMatrixParameterName, "The distance matrix of the TSP."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestIndividualQualityAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      var schemata = (ItemArray<IntArray>)Results[SchemataParameterName].Value;

      var schemataMatrix = new Dictionary<IntArray, string>(new SchemaEqualityComparer());

      foreach (var schema in schemata) {
        if (schema == schemata.Last()) {
          break; // the last one is the best found solution
        }

        if (!schemataMatrix.ContainsKey(schema)) {
          string data = ExploreSchemaOccurrence(graph, schema, schemata.Last());
          schemataMatrix.Add(schema, data);
        }
      }

      // order schemas according to their global occurrence count
      var sortedSchemataMatrix = schemataMatrix.OrderByDescending(x => x.Value[0]).ToDictionary(x => x.Key, x => x.Value);

      var stringMatrix = new StringMatrix(sortedSchemataMatrix.Keys.Count, 2);
      for (int i = 0; i < sortedSchemataMatrix.Keys.Count; i++) {
        var element = sortedSchemataMatrix.ElementAt(i);
        stringMatrix[i, 0] = new Permutation(PermutationTypes.RelativeUndirected, element.Key).ToString();
        stringMatrix[i, 1] = element.Value;
      }
      stringMatrix.ColumnNames = new[] { "Schema", "Occurence in hierarchy" };

      Results.Add(new Result(SchemaQualityMatrixParameterName, stringMatrix));

      return base.Apply();
    }

    private string ExploreSchemaOccurrence(GenealogyGraph<Permutation> graph, IntArray schema, IntArray parent) {
      SchemaEqualityComparer comparer = new SchemaEqualityComparer();

      var lastGen = graph.Values.Where(x => x.Rank == GenerationsParameter.ActualValue.Value);
      var parents = lastGen.Where(x => comparer.Equals((IntArray)x.Data, parent));
      var ancestors = parents.First().Ancestors().Where(x => AlgorithmBehaviorHelpers.IsSubtour(schema, (Permutation)x.Data));

      StringBuilder sb = new StringBuilder();
      foreach (var a in ancestors) {
        sb.Append(a.Rank + ", ");
      }
      return sb.ToString();
    }
  }
}
