﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior {
  using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

  /// <summary>
  /// An operator that analyzes schemata.
  /// </summary>
  [Item("SchemataCreator", "An operator that builds schemata.")]
  [StorableClass]
  public sealed class SchemataCreator : SingleSuccessorOperator, IAnalyzer {
    private const string GlobalTraceMapParameterName = "GlobalTraceMap";
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemaMatrixParameterName = "SchemaMatrix";
    private const string SchemataParameterName = "Schemata";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (ILookupParameter<TraceMapType>)Parameters[GlobalTraceMapParameterName]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    #endregion

    #region Properties
    public TraceMapType GlobalTraceMap {
      get { return GlobalTraceMapParameter.ActualValue; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private SchemataCreator(bool deserializing) : base(deserializing) { }
    private SchemataCreator(SchemataCreator original, Cloner cloner) : base(original, cloner) { }
    public SchemataCreator()
      : base() {
      Parameters.Add(new LookupParameter<TraceMapType>(GlobalTraceMapParameterName, "A global cache containing the whole genealogy."));
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SchemataCreator(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      var generationZero = graph.Values.Where(x => x.Rank == 0);

      // extract all subtours for each individual in generation zero
      var subtours = new Dictionary<Permutation, ItemArray<IntArray>>();
      foreach (var individual in generationZero) {
        subtours[(Permutation)individual.Data] = new ItemArray<IntArray>(AlgorithmBehaviorHelpers.ExtractSubtours((Permutation)individual.Data));
      }

      // process all schemata
      var schemataMatrix = new Dictionary<IntArray, int>(new SchemaEqualityComparer());
      foreach (var entry in subtours) {
        foreach (var schema in entry.Value) {
          if (!schemataMatrix.ContainsKey(schema)) {
            schemataMatrix.Add(schema, ExploreGlobalSchemaOccurrence(graph, schema));
          }
        }
      }

      // order schemas according to their global occurrence count
      var sortedSchemataMatrix = schemataMatrix.OrderByDescending(x => x.Value).ToDictionary(x => x.Key, x => x.Value);
      IList<string> rowNames = new List<string>();
      var schemataStringMatrix = new StringMatrix(sortedSchemataMatrix.Keys.Count, 1);
      for (int i = 0; i < sortedSchemataMatrix.Keys.Count; i++) {
        var element = sortedSchemataMatrix.ElementAt(i);
        rowNames.Add(new Permutation(PermutationTypes.RelativeUndirected, element.Key).ToString());
        schemataStringMatrix[i, 0] = element.Value.ToString();
      }
      schemataStringMatrix.RowNames = rowNames;
      schemataStringMatrix.ColumnNames = new[] { "GlobalSchemaOccurrence" };

      Results.Add(new Result(SchemataParameterName, new ItemArray<IntArray>(schemataMatrix.Keys)));
      Results.Add(new Result(SchemaMatrixParameterName, schemataStringMatrix));

      return base.Apply();
    }

    private int ExploreGlobalSchemaOccurrence(GenealogyGraph<Permutation> graph, IntArray schema) {
      int occurrences = 0;
      foreach (var individual in graph.Values)
        if (AlgorithmBehaviorHelpers.IsSubtour(schema, (Permutation)individual.Data)) occurrences++;
      return occurrences;
    }
  }
}
