﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes wildcards.
  /// </summary>
  [Item("WildcardsCreator", "An operator that builds wildcards.")]
  [StorableClass]
  public sealed class WildcardsCreator : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string SchemataResultName = "Schemata";
    private const string WildcardsResultName = "Wildcards";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private WildcardsCreator(bool deserializing) : base(deserializing) { }
    private WildcardsCreator(WildcardsCreator original, Cloner cloner) : base(original, cloner) { }
    public WildcardsCreator()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new WildcardsCreator(this, cloner);
    }

    public override IOperation Apply() {
      var schemata = (ItemArray<IntArray>)Results[SchemataResultName].Value;
      var wildcards = new List<IntArray>();
      foreach (var schema in schemata) {
        wildcards.AddRange(GenerateWildcardSchemata(schema));
      }

      Results.Add(new Result(WildcardsResultName, new ItemArray<IntArray>(wildcards)));

      return base.Apply();
    }

    private IList<IntArray> GenerateWildcardSchemata(IntArray schema) {
      string formatString = "{0:d" + schema.Length + "}";
      var combs = new List<bool[]>();
      string comb = string.Empty;

      // create all desired wildcard combinations
      int i = 0;
      int wildcardCount = 0; // first wildcard combination contains no wildcards
      while (wildcardCount < schema.Length) {
        wildcardCount = (comb = string.Format(formatString, Convert.ToInt32(Convert.ToString(i++, 2)))).Count(x => x == '1');
        if (wildcardCount > schema.Length / 8 && wildcardCount <= schema.Length / 2) // only include wildcards within certain limits
          combs.Add(comb.Select(x => x == '1').ToArray()); // '1' becomes true which indicates a wildcard
      }

      // create all desired wildcard instances
      var wildcardSchemata = new List<IntArray>();
      foreach (var c in combs) {
        wildcardSchemata.Add(new IntArray(schema.Select((x, index) => c[index] ? -1 : x).ToArray())); // -1 is a wildcard in the schema
      }

      return wildcardSchemata;
    }
  }
}
